import asyncio

from functools import partial
from izzylib.misc import Async

from .config import Config

from .. import __version__
from ..utils import methods


class Client:
	def __init__(self, **kwargs):
		self.cfg = Config(**kwargs)


	@Async
	async def request(self, *args, **kwargs):
		request = self.create_request(*args, **kwargs)

		return await self.send_request(request)


	@Async
	async def send_request(self, request):
		connection = asyncio.open_connection(request.domain, request.port, ssl=request.secure)
		reader, writer = await asyncio.wait_for(connection, timeout=self.cfg.timeout)

		transport = self.cfg.transport_class(reader, writer, self.cfg.timeout)

		for key, value in self.cfg.headers.items():
			request.headers.setall(key, value)

		await transport.write(request.compile())

		data = await transport.read_headers(request=False)
		return self.cfg.response_class(transport, *data)


	@Async
	async def connect(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='CONNECT')


	@Async
	async def delete(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='DELETE')


	@Async
	async def get(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='GET')


	@Async
	async def head(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='HEAD')


	@Async
	async def options(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='OPTIONS')


	@Async
	async def patch(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='PATCH')


	@Async
	async def post(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='POST')


	@Async
	async def put(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='PUT')


	@Async
	async def trace(self, *args, **kwargs):
		return await self.request(*args, **kwargs, method='TRACE')


	def create_request(self, url, body=b'', headers={}, cookies={}, method='GET', privkey=None, keyid=None):
		if (keyid and not privkey) or (not keyid and privkey):
			raise ValueError('Please provide a keyid and a privkey, not either')

		request = self.cfg.request_class(url, body, headers, cookies, method)

		if keyid and privkey:
			request.sign_headers(privkey, key_id)

		return request


	def run_request(self, request):
		return self.loop.run_until_complete(self.send_request(request))


	def read_body(self, response):
		return self.loop.run_until_complete(response.read())
