from asyncio.exceptions import TimeoutError
from functools import partial
from io import BytesIO
from izzylib.dotdict import DotDict
from izzylib.misc import Async

from ..utils import (
	Cookies,
	Headers
)

class ClientResponse:
	def __init__(self, transport, status, message, version, headers, cookies):
		self._body = BytesIO()

		self.transport = transport
		self.version = version
		self.status = status
		self.message = message
		self.headers = headers if type(headers) == Headers else Headers(headers)
		self.cookies = cookies if type(cookies) == Cookies else Cookies(cookies)


	def __enter__(self):
		return self


	def __exit__(self, *args):
		self.close()


	async def __aenter__(self):
		return self.__enter__()


	async def __aexit__(self, *args):
		return self.__exit__(*args)


	@property
	def type(self):
		return self.headers.getone('content-type')


	@type.setter
	def type(self, data):
		self.headers.setall('content-type', data)


	@property
	def length(self):
		return int(self.headers.getone('content-length', 0))


	@Async
	async def read(self, length=None):
		data = await self.transport.read(length or self.length or 8192)

		if data:
			self._body.write(data)

		return data


	@Async
	async def body(self, encoding=None):
		length = self._body.getbuffer().nbytes

		if not length or (self.length and self.length > length):
			data = await self.read(self.length - length)

		else:
			data = self._body.readall()

		return data.decode(encoding) if encoding else data


	@Async
	async def text(self, encoding='utf-8'):
		return await self.body(encoding)


	@Async
	async def dict(self):
		return DotDict(await self.text())


	def close(self):
		self._body.close()
