import asyncio
import signal
import socket
import sys
import time
import traceback

from functools import partial
from http_router import Router, MethodNotAllowed, NotFound
from izzylib import DotDict, Path, convert_to_boolean, logging, signal_handler
from jinja2.exceptions import TemplateNotFound
from threading import Event, Thread

from . import http_methods, error
from .config import Config
from .response import ServerResponse
#from .router import Router
from .view import Static, Manifest, Robots, Style

from ..exceptions import MethodNotHandledException, NoBlueprintForPath
from ..utils import AsyncTransport
from ..template import Template

try:
	from izzylib import Database
except ImportError:
	Database = NotImplementedError('Failed to import SQL database class')


class ApplicationBase:
	def __init__(self, appname='default', views=[], middleware=[], dbtype=None, dbargs={}, dbclass=Database, **kwargs):
		self.ctx = DotDict()
		self.name = appname
		self.cfg = Config(**kwargs)
		self.db = None
		self.router = Router(trim_last_slash=True)
		self.middleware = DotDict({'request': [], 'response': []})
		self.routes = {}

		for view in views:
			self.add_view(view)

		for mw in middleware:
			self.add_middleware(mw)

		if dbtype or dbargs:
			if isinstance(Database, Exception):
				raise Database from None

			self.db = dbclass(dbtype, **dbargs, app=self)


	def __getitem__(self, key):
		return self.ctx[key]


	def __setitem__(self, key, value):
		self.ctx[key] = value


	def __delitem__(self, key):
		del self.ctx[key]


	def get_route(self, path, method='GET'):
		return self.router(str(path), method.upper())


	def add_route(self, handler, path, method='GET'):
		self.router.bind(handler, path, methods=[method.upper()])
		self.routes[f'{method.upper()}:{path}'] = handler


	def compare_routes(self, route, path, method='GET'):
		try:
			return self.get_route(path, method) == self.routes[f'{method.upper()}:{route}']
		except:
			return False


	async def run_handler(self, request, response, path, method=None, **kwargs):
		handler = self.get_route(path, method or request.method)
		return await handler.target(request, response, **kwargs)


	def add_view(self, view):
		paths = view.__path__ if isinstance(view.__path__, list) else [view.__path__]

		view_class = view(self)

		for path in paths:
			for method in http_methods:
				try:
					self.add_route(view_class.get_handler(method), path, method)

				except MethodNotHandledException:
					pass


	def add_static(self, path, src):
		if Path(src).isdir:
			path = Path(path).join('{path:.*}')

		self.add_route(Static(src), path)


	def add_middleware(self, handler, attach=None):
		if not asyncio.iscoroutinefunction(handler):
			raise TypeError('Middleware handler must be a coroutine function or method')

		if not attach:
			try:
				arg_len = handler.__code__.co_argcount

				if arg_len == 1:
					attach = 'request'

				elif arg_len == 2:
					attach = 'response'

				else:
					raise TypeError(f'Middleware handler must have 1 (request) or 2 (response) arguments, not {arg_len}')

			except Exception as e:
				raise e from None

		assert attach in ['request', 'response']
		mwlist = self.middleware[attach]

		if handler in mwlist:
			return logging.error(f'Middleware handler already added to {attach}: {handler}')

		mwlist.append(handler)


	async def handle_request(self, request, response, path=None):
		if self.cfg.check_host:
			if request.host not in self.cfg.hosts and not request.path.startswith('/framework'):
				raise error.BadRequest(f'Host not handled on this server: {request.host}')

		handler = self.get_route(path or request.path, request.method)
		request._params = handler.params

		await self.handle_middleware(request)

		if handler.params:
			handler_response = await handler.target(request, response, **handler.params)

		else:
			handler_response = await handler.target(request, response)

		if isinstance(handler_response, dict):
			response = self.cfg.response_class(**handler_response)

		elif isinstance(handler_response, ServerResponse):
			response = handler_response

		elif handler_response:
			raise error.InternalServerError(f'Invalid server response type: {type(handler_response).__name__}')

		await self.handle_middleware(request, response)

		return response


	async def handle_middleware(self, request, response=None):
		for middleware in self.middleware['response' if response else 'request']:
			if response:
				await middleware(request, response)

			else:
				await middleware(request)


class Application(ApplicationBase):
	def __init__(self, loop=None, **kwargs):
		super().__init__(**kwargs)

		if not loop:
			try:
				loop = asyncio.get_running_loop()

			except RuntimeError:
				loop = asyncio.new_event_loop()

		self.loop = loop
		self.template = None
		asyncio.set_event_loop(self.loop)

		self.client = self.cfg.client_class(loop, *self.cfg.client_args, **self.cfg.client_kwargs)

		self._running = Event()
		self._blueprints = {}
		self._server = None
		self._tasks = []
		self._callbacks = DotDict(startup = [], shutdown = [])

		if self.cfg.tpl_default:
			if type(Template) == NotImplementedError:
				raise Template

			self.template = Template(
				self.cfg.tpl_search,
				self.cfg.tpl_globals,
				self.cfg.tpl_context,
				self.cfg.tpl_autoescape
			)

			self.template.add_env('app', self)
			self.template.add_env('cfg', self.cfg)
			self.template.add_env('len', len)

			self.template.add_search_path(self.cfg.frontend_path)
			self.add_view(Manifest)
			#self.add_view(Robots)
			self.add_view(Style)
			self.add_static('/framework/static/', self.cfg.frontend_path.join('static'))


	def __repr__(self):
		return f'Application({self.name}, listen={self.cfg.listen}, port={self.cfg.port})'


	@property
	def running(self):
		return self._running.is_set()


	@running.setter
	def running(self, value):
		if convert_to_boolean(value):
			self._running.set()

		else:
			self._running.clear()


	def add_task(self, state, callback, *args, **kwargs):
		assert state in ['startup', 'shutdown']
		assert asyncio.iscoroutinefunction(callback)
		self._callbacks[state].append((callback, args, kwargs))


	def add_startup_task(self, callback, *args, **kwargs):
		self.add_task('startup', callback, *args, **kwargs)


	def add_shutdown_task(self, callback, *args, **kwargs):
		self.add_task('shutdown', callback, *args, **kwargs)


	def remove_task(self, state, callback):
		assert state in ['startup', 'shutdown']

		for task_callback, _, _ in self._callbacks[state]:
			if task_callback == task:
				self._callbacks[state].remove(task)


	def remove_startup_task(self, callback):
		self.remove_task('startup', callback)


	def remove_shutdown_task(self, callback):
		self.remove_task('shutdown', callback)


	def add_blueprint(self, bp):
		assert bp.prefix not in self._blueprints.values()

		self._blueprints[bp.prefix] = bp


	def get_blueprint_for_path(self, path):
		for bppath, bp in self._blueprints.items():
			if path.startswith(bppath):
				return bp

		raise NoBlueprintForPath(path)


	def render(self, *args, **kwargs):
		return self.template.render(*args, **kwargs)


	def run(self):
		if self._server:
			return

		signal_handler(self.stop)
		return asyncio.run(self.handle_run_server())


	def start(self):
		if self._server:
			return

		signal_handler(self.stop)
		return asyncio.ensure_future(self.handle_run_server())


	def stop(self, *_):
		if not self._server:
			print('server not running')
			return

		self.running = False

		if self.cfg.sig_handler:
			self.cfg.sig_handler(self, *self.cfg.sig_handler_args, **self.cfg.sig_handler_kwargs)


	async def handle_run_server(self):
		## Create _server object
		if self.cfg.socket:
			logging.info(f'Starting server on {self.cfg.socket}')

			self._server = await asyncio.start_unix_server(
				self.handle_client,
				path = self.cfg.socket
			)

		else:
			logging.info(f'Starting server on {self.cfg.listen}:{self.cfg.port}')

			self._server = await asyncio.start_server(
				self.handle_client,
				host = self.cfg.listen,
				port = self.cfg.port,
				family = socket.AF_INET,
				reuse_address = True,
				reuse_port = True
			)

		self.running = True

		## Run startup tasks
		for callback, args, kwargs in self._callbacks['startup']:
			self._tasks.append(asyncio.ensure_future(callback(self, *args, **kwargs)))

		## Wait for server to finish
		try:
			while self._server.is_serving() and self.running:
				await asyncio.sleep(0.1)

			self._server.close()
			await self._server.wait_closed()

		except:
			traceback.print_exc()

		self.running = False
		signal_handler(None)

		## Run shutdown tasks
		for callback, args, kwargs in self._callbacks['shutdown']:
			try:
				await asyncio.wait_for(callback(self, *args, **kwargs), 10)

			except TimeoutError:
				pass

		for task in self._tasks:
			task.cancel()

			try:
				await task

			except asyncio.CancelledError:
				pass

		self._tasks = []
		self._server = None
		logging.info('Server stopped')


	async def handle_client(self, reader, writer):
		transport = AsyncTransport(reader, writer, self.cfg.timeout)
		request = None
		response = None

		try:
			request = self.cfg.request_class(self, transport)
			response = self.cfg.response_class(request=request)

			try:
				await request.parse_headers()

			except asyncio.exceptions.IncompleteReadError as e:
				request = None
				raise e from None

			try:
				# this doesn't work all the time for some reason
				blueprint = self.get_blueprint_for_path(request.path)
				response = await blueprint.handle_request(request, response, blueprint.prefix)

			except NoBlueprintForPath:
				response = await self.handle_request(request, response)

			#except Exception as e:
				#traceback.print_exc()

		except NotFound:
			response = self.cfg.response_class(request=request).set_error('Not Found', 404)

		except MethodNotAllowed:
			response = self.cfg.response_class(request=request).set_error('Method Not Allowed', 405)

		except error.RedirError as e:
			response = self.cfg.response_class.new_redir(e.path, e.status)

		except error.HttpError as e:
			response = self.cfg.response_class(request=request).set_error(e.message, e.status)

		except TemplateNotFound as e:
			response = self.cfg.response_class(request=request).set_error(f'Template not found: {e}', 500)

		except:
			traceback.print_exc()

			if not response.streaming:
				## Don't use a custom response class here just in case it caused the error
				response = ServerResponse(request=request).set_error('Server Error', 500)

		if not response.streaming:
			try:
				response.headers.update(self.cfg.default_headers)
				await transport.write(response.compile())

				if self.cfg.access_log and request and request.log and not request.path.startswith('/framework'):
					logging.info(f'{request.remote} {request.method} {request.path} {response.status} {len(response.body)} {request.agent}')

			except:
				traceback.print_exc()

		await transport.close()


class Blueprint(ApplicationBase):
	def __init__(self, prefix, **kwargs):
		super().__init__(**kwargs)

		self.prefix = prefix


## might keep this
def set_response(request, resp_class, func, *args, **kwargs):
	try:
		return getattr(resp_class, func)(*args, **kwargs)
	except:
		traceback.print_exc()
		return ServerResponse(request=request).set_error('Server Error', 500)
