import asyncio, email, traceback

from datetime import datetime, timezone
from izzylib import DotDict, MultiDotDict, Url, Address
from urllib.parse import unquote_plus

from .misc import Cookies, Headers, CookieItem

try: from ..signatures import verify_headers
except ImportError: verify_headers = None


UtcTime = timezone.utc
LocalTime = datetime.now(UtcTime).astimezone().tzinfo


class ServerRequest:
	__slots__ = [
		'_body', '_form', '_method', '_app', '_params',
		'address', 'path', 'version', 'headers', 'cookies',
		'query', 'raw_query', 'transport', 'log'
	]

	ctx = DotDict()

	def __init__(self, app, transport):
		super().__init__()

		self._app = app
		self._body = b''
		self._form = DotDict()
		self._method = None
		self._params = None

		self.transport = transport
		self.headers = Headers()
		self.cookies = Cookies()
		self.query = DotDict()

		try: self.address = Address(transport.client_address)
		except: self.address = None
		self.path = None
		self.version = None
		self.raw_query = None
		self.log = True

		self.verified = False


	def __getitem__(self, key):
		return self.ctx[key]


	def __setitem__(self, key, value):
		self.ctx[key] = value


	def __getattr__(self, key):
		if key in self.__slots__:
			return super().__getattribute__(self, key)

		try:
			return self.ctx[key]
		except:
			raise AttributeError(key)


	def __setattr__(self, key, value):
		try:
			super().__setattr__(key, value)

		except:
			self.ctx[key] = value


	@property
	def app(self):
		return self._app or get_app()


	@app.setter
	def app(self, app):
		self._app = app


	@property
	def agent(self):
		return self.headers.getone('User-Agent', 'no agent')


	@property
	def accept(self):
		return self.headers.getone('Accept', '')


	@property
	def content_type(self):
		return self.headers.getone('Content-Type', '')


	@property
	def date(self):
		date_str = self.headers.getone('Date')

		if date_str:
			date = datetime.strptime(date_str, '%a, %d %b %Y %H:%M:%S GMT')
			date = date.replace(tzinfo=UtcTime)
			return date.astimezone(LocalTime)

		# not sure if this should stay
		return datetime.now(LocalTime)


	@property
	def host(self):
		return self.headers.getone('Host')


	@property
	def length(self):
		return int(self.headers.getone('Content-Length', 0))


	@property
	def method(self):
		return self._method


	@method.setter
	def method(self, data):
		self._method = data.upper()


	@property
	def params(self):
		return self._params


	@property
	def protocol(self):
		return self.headers.getone('X-Forwarded-Proto', 'http')


	@property
	def remote(self):
		try:
			return Address(self.headers.getone('X-Real-Ip', self.headers.getone('X-Forwarded-For', self.address)))

		except:
			return None


	async def body(self):
		if not self._body and self.length:
			self._body = await self.transport.read(self.length)

		return self._body


	async def text(self):
		return (await self.body()).decode('utf-8')


	async def dict(self):
		logging.warning('Request.dict will be removed in a future update')
		return DotDict(await self.body())


	async def json(self):
		return DotDict(await self.body())


	async def form(self):
		if not self._form and 'application/x-www-form-urlencoded' in self.content_type:
			for line in unquote_plus(await self.text()).split('&'):
				try: key, value = line.split('=', 1)
				except: key, value = line, None

				self._form[key] = value

		return self._form


	async def parse_headers(self):
		method, path, version, headers, cookies = await self.transport.read_headers()

		self.path = path or ''
		self.version = version
		self.method = method
		self.headers = headers
		self.cookies = cookies

		self.query = Url(f'{self.protocol}://{self.host}{self.path or "/"}').query


	def new_response(self, *args, **kwargs):
		return self.app.cfg.response_class(*args, **kwargs)


	async def verify_signature(self, actor):
		if not verify_headers:
			raise ImportError('Failed to import verify_headers from izzylib.http_signatures.')

		self.verified = verify_headers(
			headers = {k: self.headers.getone(k) for k in self.headers.keys()},
			method = self.method,
			path = self.path,
			actor = actor,
			body = await self.body()
		)

		return self.verified
