import json, sys

from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from base64 import b64decode, b64encode
from datetime import datetime
from functools import lru_cache
from izzylib import DefaultDotDict, DotDict, Url, izzylog
from tldextract import extract


def generate_rsa_key():
	privkey = RSA.generate(2048)

	key = DotDict({'PRIVKEY': privkey, 'PUBKEY': privkey.publickey()})
	key.update({'privkey': key.PRIVKEY.export_key().decode(), 'pubkey': key.PUBKEY.export_key().decode()})

	return key


def parse_signature(signature: str):
	return Signature(signature)


def verify_headers(headers: dict, method: str, path: str, actor: dict, body=None):
	'''Verify a header signature

		headers: A dictionary containing all the headers from a request
		method: The HTTP method of the request
		path: The path of the HTTP request
		actor (optional): A dictionary containing the activitypub actor and the link to the pubkey used for verification
		body (optional): The body of the request. Only needed if the signature includes the digest header
		fail (optional): If set to True, raise an error instead of returning False if any step of the process fails
	'''

	headers = {k.lower(): headers[k] for k in headers}
	headers['(request-target)'] = f'{method.lower()} {path}'
	signature = Signature(headers.get('signature'))
	digest = headers.get('digest')
	missing_headers = [k for k in headers if k in ['date', 'host'] if headers.get(k) == None]

	if not signature:
		raise AssertionError('Missing signature')

	## Add digest header to missing headers list if it doesn't exist
	if method.lower() == 'post' and not digest:
		missing_headers.append('digest')

	## Fail if missing date, host or digest (if POST) headers
	if missing_headers:
		raise AssertionError(f'Missing headers: {missing_headers}')

	## Fail if body verification fails
	if digest:
		digest_hash = parse_body_digest(headers.get('digest'))

		if not verify_string(body, digest_hash.sig, digest_hash.alg):
			raise AssertionError('Failed body digest verification')

	pubkey = actor.publicKey['publicKeyPem']

	return sign_pkcs_headers(pubkey, {k:v for k,v in headers.items() if k in signature.headers}, sig=signature)


def verify_request(request, actor: dict):
	'''Verify a header signature from a SimpleASGI request

		request: The request with the headers to verify
		actor: A dictionary containing the activitypub actor and the link to the pubkey used for verification
	'''

	return verify_headers(
		headers = request.headers,
		method = request.method,
		path = request.path,
		actor = actor,
		body = request.body
	)


### Helper functions that shouldn't be used directly ###
def parse_body_digest(digest):
	if not digest:
		raise AssertionError('Empty digest')

	parsed = DotDict()
	alg, sig = digest.split('=', 1)

	parsed.sig = sig
	parsed.alg = alg.replace('-', '')

	return parsed


def sign_pkcs_headers(key: str, headers: dict, sig=None):
	if sig:
		head_items = [f'{item}: {headers[item]}' for item in sig.headers]

	else:
		head_items = [f'{k.lower()}: {v}' for k,v in headers.items()]

	head_string = '\n'.join(head_items)
	head_bytes = head_string.encode('UTF-8')

	KEY = RSA.importKey(key)
	pkcs = PKCS1_v1_5.new(KEY)
	h = SHA256.new(head_bytes)

	if sig:
		try:
			return pkcs.verify(h, b64decode(sig.signature))

		except ValueError:
			return False

	else:
		return pkcs.sign(h)


def sign_request(request, privkey, keyid):
	assert isinstance(request.body, bytes)

	request.set_header('(request-target)', f'{request.method.lower()} {request.url.path}')
	request.set_header('host', request.url.host)
	request.set_header('date', datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S GMT'))

	if request.body:
		body_hash = b64encode(SHA256.new(request.body).digest()).decode("UTF-8")
		request.set_header('digest', f'SHA-256={body_hash}')
		request.set_header('content-length', str(len(request.body)))

	sig = {
		'keyId': keyid,
		'algorithm': 'rsa-sha256',
		'headers': ' '.join([k.lower() for k in request.headers.keys()]),
		'signature': b64encode(sign_pkcs_headers(privkey, request.headers)).decode('UTF-8')
	}

	sig_items = [f'{k}="{v}"' for k,v in sig.items()]
	sig_string = ','.join(sig_items)

	request.set_header('signature', sig_string)

	request.unset_header('(request-target)')
	request.unset_header('host')

	return request


def verify_string(string, enc_string, alg='SHA256', fail=False):
	if type(string) != bytes:
		string = string.encode('UTF-8')

	body_hash = b64encode(SHA256.new(string).digest()).decode('UTF-8')

	if body_hash == enc_string:
		return True

	if fail:
		raise AssertionError('String failed validation')

	else:
		return False


class Signature(str):
	__parts = {}

	def __init__(self, signature: str):
		if not signature:
			raise AssertionError('Missing signature header')

		split_sig = signature.split(',')

		for part in split_sig:
			key, value = part.split('=', 1)
			value = value.replace('"', '')

			self.__parts[key.lower()] = value.split() if key == 'headers' else value


	def __new__(cls, signature: str):
		return str.__new__(cls, signature)


	def __new2__(cls, signature: str):
		data = str.__new__(cls, signature)
		data.__init__(signature)

		return


	def __getattr__(self, key):
		return self.__parts[key]


	@property
	def sig(self):
		return self.__parts['signature']


	@property
	def actor(self):
		return Url(self.keyid.split('#')[0])


	@property
	def domain(self):
		return self.actor.host


	@property
	def top_domain(self):
		return '.'.join(extract(self.domain)[1:])
