import codecs, traceback, os, json, xml

from functools import partial
from hamlish_jinja import HamlishExtension
from izzylib import DotDict, Path, color, izzylog
from jinja2 import Environment, FileSystemLoader, ChoiceLoader, select_autoescape
from os import listdir, makedirs
from os.path import isfile, isdir, getmtime, abspath
from xml.dom import minidom

try:
	from sanic import response as Response

except ModuleNotFoundError:
	Response = None


class Template(Environment):
	def __init__(self, search=[], global_vars={}, context=None, autoescape=True):
		self.search = FileSystemLoader([])

		super().__init__(
			loader=self.search,
			extensions=[HamlishExtension],
			lstrip_blocks=True,
			trim_blocks=True
		)

		self.autoescape = autoescape
		self.func_context = context

		self.hamlish_file_extensions=('.haml', '.jhaml')
		self.hamlish_enable_div_shortcut=True
		self.hamlish_mode = 'indented'

		for path in search:
			self.add_search_path(Path(path))

		self.globals.update({
			'cleanhtml': lambda text: ''.join(xml.etree.ElementTree.fromstring(text).itertext()),
			'color': color.Color,
			'lighten': color.lighten,
			'darken': color.darken,
			'saturate': color.saturate,
			'desaturate': color.desaturate,
			'rgba': color.rgba
		})

		self.globals.update(global_vars)


	def add_search_path(self, path, index=None):
		if not path.exists:
			raise FileNotFoundError(f'Cannot find search path: {path}')

		if path not in self.search.searchpath:
			loader = os.fspath(path)

			if index != None:
				self.search.searchpath.insert(index, loader)
			else:
				self.search.searchpath.append(loader)


	def set_context(self, context):
		if not hasattr(context, '__call__'):
			izzylog.error('Context is not callable')
			return

		if not isinstance(context({}), dict):
			izzylog.error('Context does not return a dict or dict-like object')
			return

		self.func_context = context


	def add_env(self, k, v):
		self.globals[k] = v


	def del_env(self, var):
		if not self.globals.get(var):
			raise ValueError(f'"{var}" not in global variables')

		del self.var[var]


	def update_env(self, data):
		if not isinstance(data, dict):
			raise ValueError(f'Environment data not a dict')

		self.globals.update(data)


	def add_filter(self, funct, name=None):
		name = funct.__name__ if not name else name
		self.filters[name] = funct


	def del_filter(self, name):
		if not self.filters.get(name):
			raise valueError(f'"{name}" not in global filters')

		del self.filters[name]


	def update_filter(self, data):
		if not isinstance(data, dict):
			raise ValueError(f'Filter data not a dict')

		self.filters.update(data)


	def render(self, tplfile, context_data={}, headers={}, cookies={}, request=None, pprint=False):
		if not isinstance(context_data, dict):
			raise TypeError(f'context for {tplfile} not a dict: {type(context)} {context}')

		context = DotDict(self.globals)
		context.update(context_data)
		context['request'] = request if request else {'headers': headers, 'cookies': cookies}

		if self.func_context:
			# Backwards compat
			try:
				context = self.func_context(context)

			except TypeError:
				context = self.func_context(context, {})

		if context == None:
			izzylog.warning('Template context was set to "None"')
			context = {}

		result = self.get_template(tplfile).render(context)

		if pprint and any(map(tplfile.endswith, ['haml', 'html', 'xml'])):
			return minidom.parseString(result).toprettyxml(indent="  ")

		else:
			return result


	def response(self, request, tpl, ctype='text/html', status=200, **kwargs):
		if not Response:
			raise ModuleNotFoundError('Sanic is not installed')

		html = self.render(tpl, request=request, **kwargs)
		return Response.HTTPResponse(body=html, status=status, content_type=ctype, headers=kwargs.get('headers', {}))
