import re

from xml.dom import minidom
from izzylib import DotDict

__all__ = []
search_pattern = re.compile(r"{{([A-Za-z0-9_]+)}}")


def register(cls):
	__all__.append(cls.__name__)
	return cls


@register
def prettyprint(html, indent=4):
	if indent:
		if isinstance(indent, int):
			ET.indent(tree, ' ' * indent)

		elif isinstance(indent, str):
			ET.indent(tree, indent)

		else:
			raise TypeError(f'indent must be a str or int, not {indent.__class__.__name__}')

	return minidom.parseString(html).toprettyxml(indent=indent)


@register
class Environment(DotDict):
	_base = None
	_base_vars = None


	def __init__(self, **templates):
		for value in templates.values():
			assert isinstance(value, Template)

		super().__init__(templates)


	def set_base(self, template, *variables):
		self._base = base
		self._base_vars = variables


	def render(self, template, **kwargs):
		return self[template].render(**kwargs)


@register
class Template:
	def __init__(self, html, **defaults):
		self.html = html
		self.defaults = defaults


	def render(self, **kwargs):
		html = str(self.html)

		for key in search_pattern.findall(html):
			html = html.replace('{{' + key + '}}', kwargs.get(key, self.defaults.get(key, '')))

		return html


@register
class Element(DotDict):
	_body = ''
	_tag = ''
	_autoclose = False

	ATTRIBUTES = {
		'accesskey', 'autocapitalize', 'autofocus', 'class', 'contenteditable', 'contextmenu',
		'data-*', 'dir', 'draggable', 'enterkeyhint', 'exportparts', 'hidden', 'id', 'inputmode',
		'is', 'itemid', 'itemprop', 'itemref', 'itemscope', 'itemtype', 'lang', 'nonce', 'part',
		'slot', 'spellcheck', 'style', 'tabindex', 'title', 'translate'
	}


	def __init__(self, *elements, autoclose=None, tag=None, **attributes):
		super().__init__({k.replace('_', '-'): v for k, v in attributes.items()})

		self._body = ' '.join(str(e) for e in elements)

		if autoclose != None:
			self._autoclose = autoclose

		self._tag = tag or self.__class__.__name__.lower()


	def __str__(self):
		return self.build()


	@property
	def tag(self):
		return self._tag


	@tag.setter
	def tag(self, value):
		self._tag = value


	@property
	def body(self):
		return self._body


	@body.setter
	def body(self, value):
		self._body = value


	@property
	def autoclose(self):
		return self._autoclose


	@autoclose.setter
	def autoclose(self, value):
		self._autoclose = value


	def set_body(self, *elements):
		self.body = ' '.join(str(e) for e in elements)
		return self


	def build(self, indent=4):
		attr_list = []

		for key, value in self.items():
			if not value:
				attr_list.append(str(value))

			else:
				attr_list.append(f'{key}="{value}"')

		attrs = ' '.join(attr_list)

		if self.autoclose:
			html = f'<{self.tag} {attrs} />'

		else:
			html = f'<{self.tag} {attrs}>{self.body}</{self.tag}>'

		if indent:
			return prettyprint(html, indent)

		return html


@register
class Html(Element):
	ATTRIBUTES = {*Element.ATTRIBUTES,
		'manifest', 'version', 'xmlns'
	}


@register
class Head(Element):
	def favicon(self, *args, **kwargs):
		self._body += Link.favicon(*args, **kwargs)
		return self


	def manifest(self, *args, **kwargs):
		self._body += Link.manifest(*args, **kwargs)
		return self


	def stylesheet(self, *args, **kwargs):
		self._body += Link.stylesheet(*args, **kwargs)
		return self


	def title(self, title):
		self._body += Title(title)
		return self


@register
class Title(Element):
	def __init__(self, title):
		super().__init__(title)


@register
class Link(Element):
	_autoclose = True

	def __init__(self, href, rel, mimetype=None):
		if mimetype:
			super().__init__(href=href, rel=rel, type=mimetype)

		else:
			super().__init__(href=href, rel=rel)


	@classmethod
	def stylesheet(cls, href, mimetype='text/css'):
		return cls(href, 'stylesheet', mimetype)


	@classmethod
	def manifest(cls, href, mimetype='application/json'):
		return cls(href, 'manifest', mimetype)


	@classmethod
	def favicon(cls, href, mimetype='image/png'):
		return cls(href, 'shortcut icon', mimetype)


@register
class Meta(Element):
	_autoclose = True

	@classmethod
	def viewport(cls, width='device-width', initial_scale=1):
		return cls(content=f'width={width}, initial-scale={initial_scale}')


	@classmethod
	def refresh(cls, target, timeout=None):
		if timeout != None:
			content = f'{timeout};{target}'

		else:
			content = target

		return cls(http_equiv='refresh', content=content)


	@classmethod
	def charset(cls, encoding='utf-8'):
		return cls(charset=encoding)


@register
class Body(Element):
	pass


@register
class Div(Element):
	pass


@register
class Span(Element):
	pass
